/* a HyperCard XFCN that opens the chosen text file from the standard
 * file dialog, and then builds a complete inverted index to that
 * file ... uses the 'qndxr.c' method:  take the input
 * file, break it up into small chunks, and for each chunk build in
 * memory an index consisting of *.k and *.p files (respectively key
 * and ptr files) ... then merge the chunks together hierarchially
 * until all done.  When finished, return "1" to hypercard if successful;
 * beep and return a null string ("") if failure due to no disk space,
 * a bad error detected, etc.
 *
 * call it as a function:
 *
 * buildIndex (x0, y0, buffersize)
 *
 * where x0 and y0 are the horizontal and vertical offsets (from the
 * top left screen corner) for files dialogs, and where the chosen total
 * buffersize is something like 200 kB (more if you have more
 * memory available from within HyperCard).
 *
 * This version has been modified to fit the Macintosh character set.  Word
 * breaks occur at control characters and punctuation.  I've tried to be
 * friendly and not break words at any symbols that I recognize to be
 * foreign letters.  However, I do not alphabetize those symbols correctly
 * in any sense of the word!  I've eliminated the 'embedded punctuation'
 * parser option, for speed and simplicity....
 *
 * make this be XFCN number 2024 or so and call it "buildIndex"....
 *
 * 871211, 1222-24-27-29-31 ^z
 *
 * Some commentary/documentation from an earlier version follows....
 */

/*	revised main indexer program 'multindexer.c' by ^z -- 871106-...
 *  formerly 'qndxr.c', and before that, 'ndxr.c'... 
 *
 * Synopsis of the qndxr.c program's approach to sorting an index:
 *	- load a chunk of the input file into memory; during the loading,
 *		filter the characters appropriately, turning lower-case
 *		into capitals, changing delimiters into \0's, and noting down
 *		the locations of the beginnings of keys (words) in a ptr array;
 *	- do a quicksort on the ptr array, using the keys in the buffer
 *		to sort on;
 *	- write the resulting sorted list of pointers along with their keys
 *		to disk as files, named z0k0 and z0p0, in the *.k and *.p formats
 *		used in ndxr.15;
 *	- repeat the above steps with another chunk of the input file, and
 *		name the resulting files z0k1 and z0p1; repeat, etc., until the
 *		input file is all sorted into chunks;
 *	- begin to merge the sorted files in an N-way merge ... for the
 *		specific case of N=2, for example, merge z0k0/z0p0 and
 *		z0k1/z0p1 into z1k0/z1p0; merge z0k2/z0p2 and z0k3/z0p3 into
 *		z1k1/z1p1; etc., deleting the 0th-generation files as they are
 *		merged;
 *	- merge z1k0/z1p0 and z1k1/z1p1 into z2k0/z2p0, etc., and continue
 *		merging subsequent generations until everything has become a
 *		single pair of index files, znk0/znp0, which is then renamed
 *		to be the original document name with '.k' and '.p' endings.
 *
 *	---------------------------------------------------------------
 *
 * Comparison with the older radix sort approach:
 *	The new quicksort/mergesort method gains a bit in speed (since so
 *	much of the work is done in memory rather than streaming from disk
 *	back and forth) and also saves on disk space requirements (since
 *	the k and p files are significantly compressed relative to the raw
 *	index files formerly used during index sorting).  The old radix sort
 *	tended to only achieve about 2 MB/hour indexing on my Mac Plus setup,
 *	and it required 5-6 times the size of the original file in free
 *	disk space during indexing.  This new approach achieves >4 MB/hour
 *	in tests on the same Mac Plus, and only requires about 1.9 times
 * the original file size in free space!
 *
 *	The new approach also allows for greater key length -- try recompiling
 *	with KEY_LENGTH of 28, for instance -- without such severe disk space
 *	penalties as the old radix sort would have incurred, and without severe
 *	time penalties.  (Duplicate words in the chunks are only stored once
 *	in the key file, though each must have an entry in the pointer file.)
 *
 *	The only obvious disadvantage of the quicksort/mergesort approach is
 *	that it is an O(NlogN) procedure rather than O(N), and thus gets slower
 *	when files get very large.  (Radix sorting is strictly linear in N,
 *	in theory anyway.)
 *
 *	---------------------------------------------------------------
 *
 *	For further details, contact the author:
 *     Mark Zimmermann             science (at) NEMS.ARPA
 *     9511 Gwyndale Dr.           [75066,2044] on CompuServe
 *     Silver Spring, MD  20910
 *     301-565-2166
 *	---------------------------------------------------------------
 */

#include "buildIndex.1.h"
#include <EventMgr.h>

/* define my global variables here ... the table to filter
 * characters to desired form... note that I have attempted to
 * do my best in converting lower-case to capitals, but no guarantees!
 */

char filter_table[256] =
  {	   0,   0,   0,   0,   0,   0,   0,   0,
	   0,   0,   0,   0,   0,   0,   0,   0,
  	   0,   0,   0,   0,   0,   0,   0,   0,
  	   0,   0,   0,   0,   0,   0,   0,   0,
	   0,   0,   0,   0,   0,   0,   0,   0,
	   0,   0,   0,   0,   0,   0,   0,   0,
	 '0', '1', '2', '3', '4', '5', '6', '7',
	 '8', '9',   0,   0,   0,   0,   0,   0,
	   0, 'A', 'B', 'C', 'D', 'E', 'F', 'G',
	 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
	 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
	 'X', 'Y', 'Z',   0,   0,   0,   0,   0,
	   0, 'A', 'B', 'C', 'D', 'E', 'F', 'G',
	 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
	 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
	 'X', 'Y', 'Z',   0,   0,   0,   0,   0,
	0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,
	0xCB,0x89,0x80,0xCC,0x81,0x82,0x83,0x8F,
	0x90,0x91,0x92,0x93,0x94,0x95,0x84,0x97,
	0x98,0x99,0x85,0xCD,0x9C,0x9D,0x9E,0x86,
	   0,   0,   0,   0,   0,   0,   0,0xA7,
	   0,   0,   0,   0,   0,   0,0xAE,0xAF,
	   0,   0,   0,   0,   0,0xB5,0xC6,0xB7,
	0xB8,0xB8,   0,0xBB,0xBC,0xBD,0xAE,0xAF,
	   0,   0,   0,   0,   0,   0,0xC6,   0,
	   0,   0,   0,0xCB,0xCC,0xCD,0xCE,0xCE,
	   0,   0,   0,   0,   0,   0,   0,   0,
	0xD8,   0,   0,   0,   0,   0,   0,   0,
	   0,   0,   0,   0,   0,   0,   0,   0,
	   0,   0,   0,   0,   0,   0,   0,   0,
	   0,   0,   0,   0,   0,   0,   0,   0,
	   0,   0,   0,   0,   0,   0,   0,   0  } ;


pascal void main (paramPtr)
  XCmdBlockPtr paramPtr;
  {
	Str255 doc_filename;
	char info[128];
	Ptr testzbufsiz;
	WindowRecord w_record;
	WindowPtr info_window;
	Rect b_rect;
	Handle answer;
	int doc_file, vRef0, pass_number, generation_number,
		prev_gen_number, file_number, x0, y0, i;
	long zbufsiz, doc_filesize;

	SetUpA4();
	
	if (paramPtr->paramCount != 3)
	  {
	  	SysBeep (10);
	  	RestoreA4();
	  	return;
	  }
	x0 = atol (*(paramPtr->params[0]));
	y0 = atol (*(paramPtr->params[1]));
	zbufsiz = atol (*(paramPtr->params[2]));
	zbufsiz /=  (2 * NMERGE + 2);
	zbufsiz = zbufsiz - zbufsiz % (sizeof(KEY_REC) * sizeof(long));

	b_rect.top = 30 + y0;
	b_rect.left = 12 + x0;
	b_rect.bottom = 52 + y0;
	b_rect.right = 500 + x0;
	info_window = NewWindow (&w_record, &b_rect, "\p", (Boolean)1, dBoxProc,
		(WindowPtr)-1, (Boolean)0, (long)0);
	ShowWindow (info_window);
	SetPort (info_window);
	TextFont (0);
	give_msg ("\pWelcome to buildIndex by ^z!  Please choose a text file to index.");
  	
	if (zbufsiz < sizeof(KEY_REC) * sizeof(long))
	  {
	    give_msg ("\pBad choice for buffer size -- too small!!  Click mouse to exit...");
		beepWait ();
		CloseWindow (info_window);
	  	RestoreA4();
	  	return;
	  }

	if ((testzbufsiz = NewPtr (zbufsiz * (2 * NMERGE + 2))) == NULL)
	  {
	    give_msg ("\pBad choice for buffer size -- too big!!  Click mouse to exit...");
		beepWait ();
		CloseWindow (info_window);
	  	RestoreA4();
	  	return;
	  }

  	DisposPtr (testzbufsiz);

	if (! GetFileZ (&doc_filename, &vRef0, x0, y0))
		  {
			DisposeWindow (info_window);
	  		RestoreA4();
			return;
		  }
	
	if (FSOpen (doc_filename, vRef0, &doc_file) != noErr)
	  {
	    give_msg ("\pSorry, error opening the file!  Click mouse to exit...");
		beepWait ();
		CloseWindow (info_window);
	  	RestoreA4();
	  	return;
	  }
	
	give_msg ("\pBeginning to build index subfiles -- please be patient...");
	
	pass_number = 0;
	while (build_indices (zbufsiz, doc_file, vRef0, pass_number))
		++pass_number;

	give_msg ("\pBeginning merge generation #1 -- please be patient...");

	file_number = 0;
	generation_number = 0;
	do
	  {
		prev_gen_number = generation_number;
		generation_number = merge_indices (zbufsiz, doc_file, vRef0,
						file_number, generation_number, doc_filename);
		if (generation_number == prev_gen_number)
			file_number += NMERGE;
		else
			file_number = 0;
	  }
	while (generation_number >= 0);

	FSClose (doc_file);
	answer = NewHandle (2);
	**answer = '1';
	*(*answer + 1) = '\0';
	paramPtr->returnValue = answer;
	CloseWindow (info_window);
	RestoreA4();
	return;
  }


/* following routine does the standard files dialog shtick
 * to get the name of the file to use ... cribbed from the MiniEdit
 * example that comes with LSC....  put the dialog box someplace
 * reasonable ....
 */

int GetFileZ (fnp, vRefp, x0, y0)
  Str255 *fnp;
  int *vRefp, x0, y0;
  {
	SFTypeList myTypes;
	Point SFGwhere;
	SFReply reply;

	SFGwhere.v = 90 + y0;
	SFGwhere.h = 82 + x0;
	myTypes[0] = 'TEXT';
	
	SFGetFile (SFGwhere, "\p", 0L, 1, myTypes, 0L,
				&reply);
	
	if (reply.good)
	  {
		pStrCopy ((char *)reply.fName, (char *)fnp);
		*vRefp = reply.vRefNum;
		return (1);
	  }
	else return (0);
  }


/* routine to copy a pascal string from one place to another.... used in
 * above Standard Files routine.... from LSC library....
 */

void pStrCopy (p1, p2)
  register char *p1, *p2;
  {
	register int len;
	
	len = *p2++ = *p1++;
	while (--len >= 0)
		*p2++ = *p1++;
	return;
  }


/* function to convert alphabetic string to a long integer ... from LSC
 * library.... simplified to avoid using isspace() & isdigit() ....
 */

long atol (s)
  register char *s;
  {
	register char signflag = 0;
	register long r = 0;

	while ((*s == ' '))
		s++;
		
	if (*s == '-')
	  {
		signflag = 1;
		s++;
	  }
	else if (*s == '+')
	 	s++;

	while (*s >= '0' && *s <= '9') 
		r = r * 10 + (*s++ - '0');
	
	return (signflag ? -r : r);
}



/* function to convert a number into a string and put it into the chosen
 * target place ... returns the number of characters stored ...
 * based on K&R p. 60 example of itoa().... ^z ... note, NO TERMINAL
 * '\0' in the string that is built!
 */

int putNum (ans, num)
  register char *ans;
  register long num;
  {
	register int i, j;
	int s, result;
	
	i = 0;
	s = 1;
	if (num < 0)
	  {
		num = -num;
		s = -1;
	  }
	
	do
	  	ans[i++] = num % 10 + '0';
	while ((num /= 10) > 0);
	
	if (s < 0)
		ans[i++] = '-';
	result = i;
	
	for (--i, j = 0; j < i; ++j, --i)
	  {
	  	s = ans[i];
	  	ans[i] = ans[j];
	  	ans[j] = s;
	  }

	return (result);
  }


/* tiny routine to put a message into my message window.... takes in a
 * PASCAL type string and displays it nicely....
 */

void give_msg (msg)
  char *msg;
  {
	Rect b_rect;
	
	b_rect.top = b_rect.left = 0;
	b_rect.bottom = b_rect.right = 512;

	EraseRect (&b_rect);
	MoveTo (4, 15);
	DrawString (msg);
	
	return;
  }


/* function to beep and then wait for a mouse click before returning...
 * delay for 2 ticks to debounce the mouse button a bit .... flush
 * events to avoid spurious clicks hanging around....
 */

void beepWait ()
  {
  	long tickcount;

  	SysBeep (10);
  	while (! Button())
		;
	Delay (2L, &tickcount);
	while (Button())
		;
	Delay (2L, &tickcount);
	FlushEvents (everyEvent, 0);
	return;
  }



